import re
import logging
from fnmatch import fnmatch

from typing import Literal


class ItchDownloadError(Exception):
    pass


def get_int_after_marker_in_json(text: str, marker: str, key: str) -> int | None:
    """
    Many itch.io sites use a pattern like this: Most of the HTML page
    is prerendered, but certain interactive objects are handled with
    JavaScript initialized with `I.WidgetHandler({"id": 123, ...})`
    somewhere near the end of each page. Those config blocks often
    contain metadata like game/page IDs that we want to extract.
    """
    marker_line: str | None = None
    for line in reversed(text.splitlines()):
        marker_index = line.find(marker)
        if marker_index != -1:
            marker_line = line[marker_index:]
            break

    if marker_line is None:
        return None

    # Notice double-slashes in the f-string (not r-string)!
    pattern = f'\\"{key}\\":\\s?(\\d+)'

    found_ints = re.findall(pattern, marker_line)
    if len(found_ints) != 1:
        return None

    return int(found_ints[0])


def should_skip_item_by_glob(kind: Literal["File"] | Literal["URL"], item: str, glob: str) -> bool:
    if glob and not fnmatch(item, glob):
        logging.info("%s '%s' does not match the glob filter '%s', skipping", kind, item, glob)
        return True

    return False


def should_skip_item_by_regex(kind: Literal["File"] | Literal["URL"], item: str, regex: str) -> bool:
    if regex and not re.fullmatch(regex, item):
        logging.info("%s '%s' does not match the regex filter '%s', skipping", kind, item, regex)
        return True

    return False



import os
import json
import re
import logging
import urllib.parse
import zipfile
import tarfile
from typing import TypedDict, Any

from bs4 import BeautifulSoup
from requests.exceptions import HTTPError, JSONDecodeError

from tqdm import tqdm
from tqdm.contrib.concurrent import thread_map

from .api import ItchApiClient
from .utils import ItchDownloadError, get_int_after_marker_in_json, should_skip_item_by_glob, should_skip_item_by_regex
from .consts import ITCH_GAME_URL_REGEX
from .config import Settings
from .infobox import parse_infobox, InfoboxMetadata

TARGET_PATHS = {
    "site": "site.html",
    "cover": "cover",
    "metadata": "metadata.json",
    "files": "files",
    "screenshots": "screenshots",
    "hh-metadata": "game.json",
}


class DownloadResult:
    def __init__(self, url: str, success: bool, errors: list[str] | None, external_urls: list[str]) -> None:
        self.url = url
        self.success = success
        self.errors = errors or []
        self.external_urls = external_urls or []


class GameMetadata(TypedDict, total=False):
    game_id: int
    title: str
    url: str

    errors: list[str]
    external_downloads: list[str]

    author: str
    author_url: str

    cover_url: str | None
    screenshots: list[str]
    description: str | None

    rating: dict[str, float | int]
    extra: InfoboxMetadata

    created_at: str
    updated_at: str
    released_at: str
    published_at: str


class GameDownloader:
    def __init__(self, settings: Settings, keys: dict[int, str]) -> None:
        self.settings = settings
        self.download_keys = keys
        self.client = ItchApiClient(settings.api_key, settings.user_agent)

    @staticmethod
    def get_rating_json(site: BeautifulSoup) -> dict | None:
        for ldjson_node in site.find_all("script", type="application/ld+json"):
            try:
                ldjson: dict = json.loads(ldjson_node.text.strip())
                if ldjson.get("@type") == "Product":
                    return ldjson
            except json.JSONDecodeError:
                continue  # Can't do much with this...

        return None

    @staticmethod
    def get_meta(site: BeautifulSoup, **kwargs: Any) -> str | None:  # noqa: ANN401
        """Grabs <meta property="xyz" content="value"/> values."""
        node = site.find("meta", attrs=kwargs)
        if not node:
            return None

        return node.get("content")

    def get_game_id(self, url: str, site: BeautifulSoup) -> int:
        game_id: int | None = None

        try:
            # Headers: <meta name="itch:path" content="games/12345" />
            itch_path = self.get_meta(site, name="itch:path")
            if itch_path is not None:
                # Its value should be "games/12345", so:
                game_id = int(itch_path.split("/")[-1])
        except ValueError:
            pass

        if game_id is None:
            # I.ViewGame has the "id" key in its config
            for script in site.find_all("script", type="text/javascript"):
                script_src = script.text.strip()
                marker = "I.ViewGame"
                if marker in script_src:
                    game_id = get_int_after_marker_in_json(script_src, marker, "id")
                    break

        if game_id is None:
            # We have to hit the server again :(
            data_url = url.rstrip("/") + "/data.json"
            data_request = self.client.get(data_url, append_api_key=False)
            if data_request.ok:
                try:
                    game_data = data_request.json()

                    if "errors" in game_data:
                        raise ItchDownloadError(
                            f"Game data fetching failed for {url} "
                            f"(likely access restricted, see issue #16): {game_data['errors']}"
                        )

                    if "id" in game_data:
                        game_id = int(game_data["id"])
                except (ValueError, TypeError, JSONDecodeError):
                    pass

        if game_id is None:
            raise ItchDownloadError(f"Could not get the Game ID for URL: {url}")

        return game_id

    def extract_metadata(self, game_id: int, url: str, site: BeautifulSoup) -> GameMetadata:
        rating_json: dict | None = self.get_rating_json(site)
        title = rating_json.get("name") if rating_json else None

        description: str | None = self.get_meta(site, property="og:description")
        if not description:
            description = self.get_meta(site, name="description")

        screenshot_urls: list[str] = []
        screenshots_node = site.find("div", class_="screenshot_list")
        if screenshots_node:
            screenshot_urls = [a["href"] for a in screenshots_node.find_all("a")]

        metadata = GameMetadata(
            game_id=game_id,
            title=title or site.find("h1", class_="game_title").text.strip(),
            url=url,
            cover_url=self.get_meta(site, property="og:image"),
            screenshots=screenshot_urls,
            description=description,
        )

        infobox_div = site.find("div", class_="game_info_panel_widget")
        if infobox_div:
            infobox = parse_infobox(infobox_div)
            for dt in ("created_at", "updated_at", "released_at", "published_at"):
                if dt in infobox:
                    metadata[dt] = infobox[dt].isoformat()  # noqa: PyTypedDict (non-literal TypedDict keys)
                    del infobox[dt]  # noqa: PyTypedDict (non-literal TypedDict keys)

            if "author" in infobox:
                metadata["author"] = infobox["author"]["author"]
                metadata["author_url"] = infobox["author"]["author_url"]
                del infobox["author"]

            if "authors" in infobox and "author" not in metadata:
                # Some games may have multiple authors (ex. compilations).
                metadata["author"] = "Multiple authors"
                metadata["author_url"] = f"https://{urllib.parse.urlparse(url).netloc}"

            metadata["extra"] = infobox

        agg_rating = rating_json.get("aggregateRating") if rating_json else None
        if agg_rating:
            try:
                metadata["rating"] = {"average": float(agg_rating["ratingValue"]), "votes": agg_rating["ratingCount"]}
            except:  # noqa: E722 (do not use bare `except`)
                logging.exception("Could not extract the rating metadata...")
                pass  # Nope, just, don't

        return metadata

    def get_credentials(self, title: str, game_id: int) -> dict:
        credentials = {}
        if game_id in self.download_keys:
            credentials["download_key_id"] = self.download_keys[game_id]
            logging.debug("Got credentials for %s: %s", title, str(credentials))

        return credentials

    def download_file(self, url: str, download_path: str | None, credentials: dict) -> str:
        """Performs a request to download a given file, optionally saves the
        file to the provided path and returns the final URL that was downloaded."""
        try:
            # No timeouts, chunked uploads, default retry strategy, should be all good?
            with self.client.get(url, data=credentials, stream=True, guess_encoding=True) as r:
                r.raise_for_status()

                if download_path is not None:  # ...and it will be for external downloads.
                    with tqdm.wrapattr(
                        open(download_path, "wb"),
                        "write",
                        miniters=1,
                        desc=url,
                        total=int(r.headers.get("content-length", 0)),
                    ) as f:
                        for chunk in r.iter_content(chunk_size=1048576):  # 1MB chunks
                            f.write(chunk)

                return r.url
        except HTTPError as e:
            raise ItchDownloadError(f"Unrecoverable download error: {e}") from e

    def download_file_by_upload_id(self, upload_id: int, download_path: str | None, credentials: dict) -> str:
        """Performs a request to download a given upload by its ID."""
        return self.download_file(f"/uploads/{upload_id}/download", download_path, credentials)

    @staticmethod
    def get_decompressed_content_size(target_path: str | os.PathLike[str]) -> None | int:
        """For some files, Itch API returns the decompressed file size, but serves
        compressed downloads. Try to figure out the decompressed size. It may be
        a single file in the root, or a container + files in it."""
        if zipfile.is_zipfile(target_path):
            try:
                with zipfile.ZipFile(target_path) as f:
                    # Zip files contain either directories or files. The file format
                    # is compression-aware, compress_size is packed, file_size is unpacked.
                    file_infos = [i for i in f.infolist() if not i.is_dir()]
                    return None if len(file_infos) == 0 else sum(i.file_size for i in file_infos)
            except zipfile.BadZipFile:
                return None

        if tarfile.is_tarfile(target_path):
            try:
                with tarfile.open(target_path) as f:
                    # Tar files can contain any Unix "file", so regular files,
                    # directories, symlinks, devices and FIFOs are fair game...
                    # On the other hand, TAR is not compression-aware.
                    file_infos = [i for i in f.getmembers() if i.isfile()]
                    return None if len(file_infos) == 0 else sum(i.size for i in file_infos)
            except tarfile.TarError:
                return None

        return None

    def map_metadata_to_hh(self, metadata: GameMetadata, romfile: str, slug: str, cover_filename: str) -> dict:
        tags = []

        metadata_hh = {
            "$schema": "https://raw.githubusercontent.com/gbdev/database/refs/heads/master/schemas/game-schema-d5.json",
            "slug": slug,
            "title": metadata["title"],
            "screenshots": [cover_filename] + [f"screenshots/{url.split('/')[-1]}" for url in metadata["screenshots"]],
            "typetag": "game",
            "platform": os.path.splitext(romfile)[1].upper().lstrip("."),
        }

        source_code_found = False
        if "links" in metadata["extra"]:
            for link_name, link in metadata["extra"]["links"].items():
                if any(key in link_name.lower() for key in ("github", "source", "repo")):
                    metadata_hh["repository"] = link
                    source_code_found = True

        if "code_license" in metadata["extra"]:
            for value in metadata["extra"]["code_license"].keys():  # noqa: SIM118
                if source_code_found:
                    metadata_hh["gameLicense"] = value

        # tags = metadata["tags"].keys()
        tags = []
        hh_tag_list = [
            "2D",
            "Action",
            "Adventure",
            "Arcade",
            "Arena",
            "Butano",
            "Card Game",
            "Educational",
            "Fighting",
            "Interactive Fiction",
            "Open Source",
            "Platform",
            "Puzzle",
            "RPG",
            "Racing",
            "Rhythm",
            "Shooter",
            "Simulation",
            "Sports",
            "Strategy",
            "Survival",
            "Visual Novel",
            "gb-showdown-22",
            "gbajam21",
            "gbcompo21",
            "gbcompo21-shortlist",
        ]

        # Add first genre if it exists
        genre_keys = metadata.get("extra", {}).get("genre", {})
        if genre_keys:
            first_genre = next(iter(genre_keys)).lower()
            tags.append(first_genre)

        # Add all tag keys, lowercased
        tag_keys = metadata.get("extra", {}).get("tags", {})
        tags.extend(key.lower() for key in tag_keys)

        # Create a lowercase map of hh_tag_list for matching
        hh_tag_map = {tag.lower(): tag for tag in hh_tag_list}

        # Match and return tags in original casing from hh_tag_list
        metadata_hh["tags"] = [hh_tag_map[tag] for tag in set(tags) if tag in hh_tag_map]

        metadata_hh["files"] = [{
            "default": True,
            "filename": f"files/{romfile}",
            "playable": True,
        }]

        metadata_hh["developer"] = metadata["author"]

        # metadata_hh["date"] = metadata["published_at"]
        metadata_hh["website"] = metadata["url"]

        return metadata_hh

    def download(self, url: str, skip_downloaded: bool = True) -> DownloadResult:
        match = re.match(ITCH_GAME_URL_REGEX, url)
        if not match:
            return DownloadResult(url, False, [f"Game URL is invalid: {url} - please file a new issue."], [])

        author, game = match["author"], match["game"]

        download_path = os.path.join(self.settings.download_to, author, game)

        if self.settings.hh_export:
            slug = f"{author}_{game}"
            download_path = os.path.join(self.settings.download_to, slug)

        os.makedirs(download_path, exist_ok=True)

        paths: dict[str, str] = {k: os.path.join(download_path, v) for k, v in TARGET_PATHS.items()}

        if os.path.exists(paths["metadata"]) and skip_downloaded:
            # As metadata is the final file we write, all the files
            # should already be downloaded at this point.
            logging.info("Skipping already-downloaded game for URL: %s", url)
            return DownloadResult(url, True, ["Game already downloaded."], [])

        try:
            logging.info("Downloading %s", url)
            r = self.client.get(url, append_api_key=False)
            r.raise_for_status()
        except Exception as e:
            return DownloadResult(url, False, [f"Could not download the game site for {url}: {e}"], [])

        site = BeautifulSoup(r.text, features="lxml")
        try:
            game_id = self.get_game_id(url, site)
            metadata = self.extract_metadata(game_id, url, site)
            title = metadata["title"] or game
        except ItchDownloadError as e:
            return DownloadResult(url, False, [str(e)], [])

        credentials = self.get_credentials(title, game_id)
        try:
            game_uploads_req = self.client.get(f"/games/{game_id}/uploads", data=credentials, timeout=15)
            game_uploads_req.raise_for_status()
        except Exception as e:
            return DownloadResult(url, False, [f"Could not fetch game uploads for {title}: {e}"], [])

        game_uploads = game_uploads_req.json()["uploads"]
        logging.debug("Found %d upload(s): %s", len(game_uploads), str(game_uploads))

        external_urls = []
        errors = []

        try:
            os.makedirs(paths["files"], exist_ok=True)
            romfile = None
            for upload in game_uploads:
                if any(key not in upload for key in ("id", "filename", "type", "traits", "storage")):
                    errors.append(f"Upload metadata incomplete: {upload}")
                    continue

                logging.debug(upload)
                upload_id = upload["id"]
                file_name = upload["filename"]
                file_type = upload["type"]
                file_traits = upload["traits"]
                expected_size = upload.get("size")
                upload_is_external = upload["storage"] == "external"

                if self.settings.filter_files_type and file_type not in self.settings.filter_files_type:
                    logging.info("File '%s' has ignored type '%s', skipping", file_name, file_type)
                    continue

                if (
                    self.settings.filter_files_platform
                    and file_type == "default"
                    and not any(trait in self.settings.filter_files_platform for trait in file_traits)
                ):
                    # Setup for filter_files_platform is in config.py, including the trait listing.
                    logging.info("File '%s' not for requested platforms, skipping", file_name)
                    continue

                if should_skip_item_by_glob("File", file_name, self.settings.filter_files_glob):
                    continue

                if should_skip_item_by_regex("File", file_name, self.settings.filter_files_regex):
                    continue

                logging.debug(
                    "Downloading '%s' (%d), %s",
                    file_name,
                    upload_id,
                    f"{expected_size} bytes" if expected_size is not None else "unknown size",
                )

                target_path = None if upload_is_external else os.path.join(paths["files"], file_name)

                try:
                    target_url = self.download_file_by_upload_id(upload_id, target_path, credentials)
                except ItchDownloadError as e:
                    errors.append(f"Download failed for upload {upload}: {e}")
                    continue

                if upload_is_external:
                    logging.debug("Found external download URL for %s: %s", title, target_url)
                    external_urls.append(target_url)
                    continue

                try:
                    downloaded_file_stat = os.stat(target_path)
                except FileNotFoundError:
                    errors.append(f"Downloaded file not found for upload {upload}")
                    continue

                downloaded_size = downloaded_file_stat.st_size
                content_size = self.get_decompressed_content_size(target_path)

                if (
                    all(x is not None for x in (target_path, expected_size, downloaded_size))
                    and downloaded_size != expected_size
                    and content_size != expected_size
                ):
                    errors.append(
                        f"Downloaded file size is {downloaded_size} (content {content_size}), "
                        f"expected {expected_size} for upload {upload}"
                    )

                if file_name.lower().endswith((".gb", ".gbc", ".gba")):
                    romfile = file_name

            if self.settings.hh_export and romfile is None:
                for filename in os.listdir(paths["files"]):
                    if not filename.lower().endswith(".zip"):
                        continue

                    zip_path = os.path.join(paths["files"], filename)
                    try:
                        with zipfile.ZipFile(zip_path, "r") as zip_ref:
                            # Extract all files
                            zip_ref.extractall(paths["files"])  # noqa: S202

                            for zip_info in zip_ref.infolist():
                                if zip_info.filename.lower().endswith((".gb", ".gbc", ".gba")):
                                    romfile = os.path.basename(zip_info.filename)
                                    break
                        os.remove(zip_path)  # Remove the archive after extraction

                        if romfile:
                            break
                    except zipfile.BadZipFile:
                        errors.append(f"Bad zip file: {zip_path}")

            logging.debug("Done downloading files for %s", title)
        except Exception as e:
            errors.append(f"Download failed for {title}: {e}")

        metadata["errors"] = errors
        metadata["external_downloads"] = external_urls

        if len(external_urls) > 0:
            logging.warning("Game %s has external download URLs: %s", title, external_urls)

        # TODO: Mirror JS/CSS assets
        if self.settings.mirror_web:
            os.makedirs(paths["screenshots"], exist_ok=True)
            for screenshot in metadata["screenshots"]:
                if not screenshot:
                    continue

                file_name = os.path.basename(screenshot)
                try:
                    self.download_file(screenshot, os.path.join(paths["screenshots"], file_name), credentials={})
                except Exception as e:
                    errors.append(f"Screenshot download failed (this is not fatal): {e}")

        cover_url = metadata.get("cover_url")
        cover_filename = os.path.basename(paths["cover"] + os.path.splitext(cover_url)[-1])
        if cover_url:
            try:
                self.download_file(cover_url, paths["cover"] + os.path.splitext(cover_url)[-1], credentials={})
            except Exception as e:
                errors.append(f"Cover art download failed (this is not fatal): {e}")

        if not self.settings.hh_export:
            # Skip those downloads when creating an Homebrew Hub export
            with open(paths["site"], "wb") as f:
                f.write(site.prettify(encoding="utf-8"))

            with open(paths["metadata"], "w") as f:
                json.dump(metadata, f, indent=4)
        else:
            metadata_hh = self.map_metadata_to_hh(metadata, romfile, slug, cover_filename)
            with open(paths["hh-metadata"], "w") as f:
                json.dump(metadata_hh, f, indent=4)

        if len(errors) > 0:
            logging.error("Game %s has download errors: %s", title, errors)

        logging.info("Finished job %s (%s)", url, title)
        return DownloadResult(url, len(errors) == 0, errors, external_urls)


def drive_downloads(
    jobs: list[str],
    settings: Settings,
    keys: dict[int, str],
) -> None:
    downloader = GameDownloader(settings, keys)
    tqdm_args = {
        "desc": "Games",
        "unit": "game",
    }

    if settings.parallel > 1:
        results = thread_map(downloader.download, jobs, max_workers=settings.parallel, **tqdm_args)
    else:
        results = [downloader.download(job) for job in tqdm(jobs, **tqdm_args)]

    print("Download complete!")
    for result in results:
        if not result.errors and not result.external_urls:
            continue

        if result.success:
            print(f"\nNotes for {result.url}:")
        else:
            print(f"\nDownload failed for {result.url}:")

        for error in result.errors:
            print(f"- {error}")

        for ext_url in result.external_urls:
            print(f"- External download URL (download manually!): {ext_url}")



import os
import json
import logging
import platform
import argparse
from dataclasses import dataclass, fields
from typing import Any, get_type_hints

import requests

from . import __version__


@dataclass
class Settings:
    """Available settings for itch-dl. Make sure all of them
    have default values, as the config file may not exist."""

    api_key: str | None = None
    user_agent: str = f"python-requests/{requests.__version__} itch-dl/{__version__}"

    download_to: str | None = None
    mirror_web: bool = False
    urls_only: bool = False
    parallel: int = 1

    filter_files_platform: list | None = None
    filter_files_type: list | None = None
    filter_files_glob: str | None = None
    filter_files_regex: str | None = None

    filter_urls_glob: str | None = None
    filter_urls_regex: str | None = None

    hh_export: bool = False
    verbose: bool = False


def process_platform_traits(platforms: list[str]) -> list[str] | None:
    """Converts the user-friendly platform strings into itch.io upload p_traits."""
    if not platforms:
        return None

    trait_mapping = {
        "win": "p_windows",
        "lin": "p_linux",
        "mac": "p_osx",
        "osx": "p_osx",
        "darwin": "p_osx",
        "and": "p_android",
    }

    traits = set()
    for p in platforms:
        platform_trait = None
        p = p.strip().lower().removeprefix("p_")  # noqa: PLW2901

        if p.startswith("native"):
            p = platform.system().lower()  # noqa: PLW2901
            if p.endswith("bsd"):
                logging.warning("Note: Native downloads for *BSDs are not available - Linux binaries will be used.")
                p = "linux"  # noqa: PLW2901

        for key, trait in trait_mapping.items():
            if p.startswith(key):
                platform_trait = trait
                break

        if not platform_trait:
            raise ValueError(f"Platform {p} not known!")

        traits.add(platform_trait)

    return list(traits)


def create_and_get_config_path() -> str:
    """Returns the configuration directory in the appropriate
    location for the current OS. The directory may not exist."""
    system = platform.system()
    if system == "Linux":
        base_path = os.environ.get("XDG_CONFIG_HOME") or os.path.expanduser("~/.config/")
    elif system == "Darwin":
        base_path = os.path.expanduser("~/Library/Application Support/")
    elif system == "Windows":
        base_path = os.environ.get("APPDATA") or os.path.expanduser("~/AppData/Roaming/")
    else:
        raise NotImplementedError(f"Unknown platform: {system}")

    return os.path.join(base_path, "itch-dl")


def clean_config(config_data: dict[str, Any]) -> dict[str, Any]:
    cleaned_config = {}
    settings_invalid = False
    type_hints = get_type_hints(Settings)

    # Complain about invalid types, if any:
    for key, value in config_data.items():
        if not (expected_type := type_hints.get(key)):
            logging.warning("Settings contain an unknown item, ignoring: '%s'", key)
            continue

        if not isinstance(value, expected_type):
            logging.fatal("Settings.%s has invalid type '%s', expected '%s'", key, type(value), expected_type)

            # Keep iterating to look up all the bad keys:
            settings_invalid = True
            continue

        cleaned_config[key] = value

    if settings_invalid:
        logging.fatal("Settings invalid, bailing out!")
        exit(1)

    return cleaned_config


def load_config(args: argparse.Namespace, profile: str | None = None) -> Settings:
    """Loads the configuration from the file system if it exists,
    the returns a Settings object."""
    config_path = create_and_get_config_path()
    config_file_path = os.path.join(config_path, "config.json")
    profile_file_path = os.path.join(config_path, "profiles", profile or "")

    if os.path.isfile(config_file_path):
        logging.debug("Found config file: %s", config_file_path)
        with open(config_file_path) as f:
            config_data = json.load(f)
    else:
        config_data = {}

    if os.path.isfile(profile_file_path):
        logging.debug("Found profile: %s", profile_file_path)
        with open(config_file_path) as f:
            profile_data = json.load(f)

        config_data.update(profile_data)

    # All settings from the base file:
    settings = Settings(**clean_config(config_data))

    # Apply overrides from CLI args on each field in Settings:
    for field in fields(Settings):
        key = field.name
        if value := getattr(args, key):
            setattr(settings, key, value)

    # Extra handling for special settings:
    settings.filter_files_platform = process_platform_traits(settings.filter_files_platform)

    return settings



import os
import sys
import logging
import argparse

from .handlers import get_jobs_for_url_or_path, preprocess_job_urls
from .downloader import drive_downloads
from .config import Settings, load_config
from .keys import get_download_keys
from .api import ItchApiClient
from . import __version__

import requests
import lxml
import bs4

logging.basicConfig()
logging.getLogger().setLevel(logging.INFO)


def parse_args() -> argparse.Namespace:
    # fmt: off
    parser = argparse.ArgumentParser(
        description="Bulk download stuff from Itch.io. "
                    "Docs: https://github.com/DragoonAethis/itch-dl/wiki",
        epilog=(
            "Environment: "
            f"itch-dl {__version__}, "
            f"requests {requests.__version__}, "
            f"lxml {lxml.__version__}, "
            f"bs4 {bs4.__version__}"
        )
    )

    parser.add_argument("url_or_path",
                        help="itch.io URL or path to a game jam entries.json file")
    parser.add_argument("--profile", metavar="profile", default=None,
                        help="configuration profile to load")

    # These args must match config.py -> Settings class. Make sure all defaults here
    # evaluate to False, or load_config will override profile settings.
    parser.add_argument("--api-key", metavar="key", default=None,
                        help="itch.io API key - https://itch.io/user/settings/api-keys")
    parser.add_argument("--user-agent", metavar="agent", default=None,
                        help="user agent to use when sending HTTP requests")
    parser.add_argument("--download-to", metavar="path", default=None,
                        help="directory to save results into (default: current working dir)")
    parser.add_argument("--mirror-web", action="store_true",
                        help="try to fetch assets on game sites")
    parser.add_argument("--urls-only", action="store_true",
                        help="print scraped game URLs without downloading them")
    parser.add_argument("--parallel", metavar="parallel", type=int, default=None,
                        help="how many threads to use for downloading games (default: 1)")

    parser.add_argument("--filter-files-platform", metavar="platforms", action="extend", nargs="+",
                        help="filter downloaded files by platform (windows, mac, linux, android, native), "
                             "affects only executables")
    parser.add_argument("--filter-files-type", metavar="types", action="extend", nargs="+",
                        help="filter downloaded files by type (see wiki for valid values)")
    parser.add_argument("--filter-files-glob", metavar="glob", default=None,
                        help="filter downloaded files with a shell-style glob/fnmatch (unmatched files are skipped)")
    parser.add_argument("--filter-files-regex", metavar="regex", default=None,
                        help="filter downloaded files with a Python regex (unmatched files are skipped)")

    parser.add_argument("--filter-urls-glob", metavar="glob", default=None,
                        help="filter itch URLs with a shell-style glob/fnmatch (unmatched URLs are skipped)")
    parser.add_argument("--filter-urls-regex", metavar="regex", default=None,
                        help="filter itch URLs with a Python regex (unmatched URLs are skipped)")

    parser.add_argument("--hh-export", action="store_true",
                        help="Export Homebrew Hub compliant entries")
    parser.add_argument("--verbose", action="store_true",
                        help="print verbose logs")

    return parser.parse_args()
    # fmt: on


def run() -> int:
    args = parse_args()
    if args.verbose:
        logging.getLogger().setLevel(logging.DEBUG)

    settings: Settings = load_config(args, profile=args.profile)
    if settings.verbose:
        logging.getLogger().setLevel(logging.DEBUG)

    if not settings.api_key:
        sys.exit(
            "You did not provide an API key which itch-dl requires.\n"
            "See https://github.com/DragoonAethis/itch-dl/wiki/API-Keys for more info."
        )

    url_or_path = args.url_or_path
    del args  # Do not use `args` beyond this point.

    # Check API key validity:
    client = ItchApiClient(settings.api_key, settings.user_agent)
    profile_req = client.get("/profile")
    if not profile_req.ok:
        sys.exit(
            f"Provided API key appears to be invalid: {profile_req.text}\n"
            "See https://github.com/DragoonAethis/itch-dl/wiki/API-Keys for more info."
        )

    jobs = get_jobs_for_url_or_path(url_or_path, settings)
    logging.info("Found %d URL(s) total.", len(jobs))

    jobs = preprocess_job_urls(jobs, settings)
    logging.info("Will process %d URL(s) after filtering and deduplication.", len(jobs))

    if len(jobs) == 0:
        sys.exit("No URLs to download.")

    if settings.urls_only:
        for job in jobs:
            print(job)

        return 0

    # If the download dir is not set, use the current working dir:
    settings.download_to = os.path.normpath(settings.download_to or os.getcwd())
    os.makedirs(settings.download_to, exist_ok=True)

    # Grab all the download keys (there's no way to fetch them per title...):
    keys = get_download_keys(client)

    drive_downloads(jobs, settings, keys)
    return 0



import logging

from .api import ItchApiClient

KEYS_CACHED: bool = False
DOWNLOAD_KEYS: dict[int, str] = {}
GAME_URLS: list[str] = []


def load_keys_and_urls(client: ItchApiClient) -> None:
    global KEYS_CACHED  # noqa: PLW0603 (whatever, I'll move all this to a class one day)
    logging.info("Fetching all download keys...")
    page = 1

    while True:
        logging.info("Downloading page %d (found %d keys total)", page, len(DOWNLOAD_KEYS))
        r = client.get("/profile/owned-keys", data={"page": page}, timeout=15)
        if not r.ok:
            break

        data = r.json()
        if "owned_keys" not in data:
            break  # Assuming we're out of keys already...

        for key in data["owned_keys"]:
            DOWNLOAD_KEYS[key["game_id"]] = key["id"]
            GAME_URLS.append(key["game"]["url"])

        if len(data["owned_keys"]) == data["per_page"]:
            page += 1
        else:
            break

    logging.info("Fetched %d download keys.", len(DOWNLOAD_KEYS))
    KEYS_CACHED = True


def get_owned_keys(client: ItchApiClient) -> tuple[dict[int, str], list[str]]:
    if not KEYS_CACHED:
        load_keys_and_urls(client)

    return DOWNLOAD_KEYS, GAME_URLS


def get_download_keys(client: ItchApiClient) -> dict[int, str]:
    if not KEYS_CACHED:
        load_keys_and_urls(client)

    return DOWNLOAD_KEYS


def get_owned_games(client: ItchApiClient) -> list[str]:
    if not KEYS_CACHED:
        load_keys_and_urls(client)

    return GAME_URLS



from typing import Any

import requests
from requests import Session
from urllib3.util.retry import Retry
from requests.adapters import HTTPAdapter

from .consts import ITCH_API


class ItchApiClient:
    def __init__(self, api_key: str, user_agent: str, base_url: str | None = None) -> None:
        self.base_url = base_url or ITCH_API
        self.api_key = api_key

        self.requests = Session()
        self.requests.headers["User-Agent"] = user_agent

        retry_strategy = Retry(
            total=5,
            backoff_factor=10,
            allowed_methods=["HEAD", "GET"],
            status_forcelist=[429, 500, 502, 503, 504],
        )

        # No timeouts - set them explicitly on API calls below!
        adapter = HTTPAdapter(max_retries=retry_strategy)
        self.requests.mount("https://", adapter)
        self.requests.mount("http://", adapter)

    def get(
        self,
        endpoint: str,
        append_api_key: bool = True,
        guess_encoding: bool = False,
        **kwargs: Any,  # noqa: ANN401
    ) -> requests.Response:
        """Wrapper around `requests.get`.

        :param endpoint: Path to fetch on the specified base URL.
        :param append_api_key: Send an authenticated API request.
        :param guess_encoding: Let requests guess the response encoding.
        """
        if append_api_key:
            params = kwargs.get("data") or {}

            if "api_key" not in params:
                params["api_key"] = self.api_key

            kwargs["data"] = params

        url = endpoint if endpoint.startswith("https://") else self.base_url + endpoint
        r = self.requests.get(url, **kwargs)

        # Itch always returns UTF-8 pages and API responses. Force
        # UTF-8 everywhere, except for binary file downloads.
        if not guess_encoding:
            r.encoding = "utf-8"

        return r
